/* Copyright (c) 2009 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is confidential property of Nordic
 * Semiconductor ASA.Terms and conditions of usage are described in detail
 * in NORDIC SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */
 
#include "hal_twi.h"

#include <stdbool.h>
#include <stdint.h>

#include "nrf.h"
#include "nrf_error.h"
#include "nrf_delay.h"
#include "nrf_gpio.h"
#include "nrf_sdm.h"
#include "nrf_soc.h"

/* Max cycles approximately to wait on RXDREADY and TXDREADY event, this is optimum way instead of using timers, this is not power aware, negetive side is this is not power aware */
#define MAX_TIMEOUT_LOOPS             (10000UL)        /*!< MAX while loops to wait for RXD/TXD event */

/* These macros are needed to see if the slave is stuck and we as master send dummy clock cycles to end its wait */
#define TWI_SCL_HIGH()   do { NRF_GPIO->OUTSET = (1UL << s_scl_pin); } while(0)   /*!< Pulls SCL line high */
#define TWI_SCL_LOW()    do { NRF_GPIO->OUTCLR = (1UL << s_scl_pin); } while(0)   /*!< Pulls SCL line low  */
#define TWI_SDA_HIGH()   do { NRF_GPIO->OUTSET = (1UL << s_sda_pin);  } while(0)   /*!< Pulls SDA line high */
#define TWI_SDA_LOW()    do { NRF_GPIO->OUTCLR = (1UL << s_sda_pin);  } while(0)   /*!< Pulls SDA line low  */
#define TWI_SDA_INPUT()  do { NRF_GPIO->DIRCLR = (1UL << s_sda_pin);  } while(0)   /*!< Configures SDA pin as input  */
#define TWI_SDA_OUTPUT() do { NRF_GPIO->DIRSET = (1UL << s_sda_pin);  } while(0)   /*!< Configures SDA pin as output */
#define TWI_SCL_OUTPUT() do { NRF_GPIO->DIRSET = (1UL << s_scl_pin); } while(0)   /*!< Configures SCL pin as output */

#define TWI_SDA_READ() ((NRF_GPIO->IN >> s_sda_pin) & 0x1UL)                     /*!< Reads current state of SDA */
#define TWI_SCL_READ() ((NRF_GPIO->IN >> s_scl_pin) & 0x1UL)                    /*!< Reads current state of SCL */

#define TWI_DELAY() nrf_delay_us(4) /*!< Time to wait when pin states are changed. For fast-mode the delay can be zero and for standard-mode 4 us delay is sufficient. */

static uint8_t s_scl_pin = 255;
static uint8_t s_sda_pin = 255;
static uint8_t s_ppi_chn = 255;

__inline static bool hal_twi_write(uint8_t *data, uint8_t data_length, bool issue_stop_condition)
{
    uint32_t timeout = MAX_TIMEOUT_LOOPS;   /* max loops to wait for EVENTS_TXDSENT event*/

    if(data_length == 0)
    {
        /* gently return false for requesting data of size 0 */
        return false;
    }

    NRF_TWI1->TXD = *data++;
    NRF_TWI1->TASKS_STARTTX = 1;
    while(true)
    {
        while(NRF_TWI1->EVENTS_TXDSENT == 0 && (--timeout))
        {
        }

        if(timeout == 0)
        {
            /* timeout before receiving event*/
            return false;
        }

        NRF_TWI1->EVENTS_TXDSENT = 0;
        if (--data_length == 0)
            break;

        NRF_TWI1->TXD = *data++;
    }
    if (issue_stop_condition)
    {
        timeout = MAX_TIMEOUT_LOOPS;
        
        NRF_TWI1->TASKS_STOP = 1;

        /* wait until stop sequence is sent and clear the EVENTS_STOPPED */
        while(NRF_TWI1->EVENTS_STOPPED == 0 && --timeout != 0)
        {
        }
        NRF_TWI1->EVENTS_STOPPED = 0;
    }

    return timeout != 0;
}

static __inline bool hal_twi_read(uint8_t *data, uint8_t data_length, bool issue_stop_condition)
{
    uint32_t err_code;
    uint32_t timeout            = MAX_TIMEOUT_LOOPS; /* max loops to wait for RXDREADY event*/
    uint8_t  softdevice_enabled = false;

    if(data_length == 0)
    {
        /* gently return false for requesting data of size 0 */
        return false;
    }
    
    err_code = sd_softdevice_is_enabled(&softdevice_enabled);
    if (err_code != NRF_SUCCESS)
    {
        return err_code;
    }    

    if (data_length == 1)
    {
        if (softdevice_enabled != 0)
        {
            err_code = sd_ppi_channel_assign(s_ppi_chn, &NRF_TWI1->EVENTS_BB, &NRF_TWI1->TASKS_STOP);
            if (err_code != NRF_SUCCESS)
            {
                return false;
            }
        }
        else
        {
            NRF_PPI->CH[s_ppi_chn].EEP = (uint32_t) &NRF_TWI1->EVENTS_BB;
            NRF_PPI->CH[s_ppi_chn].TEP = (uint32_t) &NRF_TWI1->TASKS_STOP;
        }
    }
    else
    {
        if (softdevice_enabled != 0)
        {
            err_code = sd_ppi_channel_assign(s_ppi_chn, &NRF_TWI1->EVENTS_BB, &NRF_TWI1->TASKS_SUSPEND);
            if (err_code != NRF_SUCCESS)
            {
                return false;
            }
        }
        else
        {
            NRF_PPI->CH[s_ppi_chn].EEP = (uint32_t) &NRF_TWI1->EVENTS_BB;
            NRF_PPI->CH[s_ppi_chn].TEP = (uint32_t) &NRF_TWI1->TASKS_SUSPEND;
        }        
    }
    if (softdevice_enabled != 0)
    {
        err_code = sd_ppi_channel_enable_set(1 << s_ppi_chn);
        if (err_code != NRF_SUCCESS)
        {
            return false;
        }
    }
    else
    {
        NRF_PPI->CHENSET = (1 << s_ppi_chn);
    }    
    
    NRF_TWI1->TASKS_STARTRX = 1;
    while(true)
    {
        timeout = MAX_TIMEOUT_LOOPS;
        while((NRF_TWI1->EVENTS_RXDREADY == 0) && (--timeout))
        {
        }

        NRF_TWI1->EVENTS_RXDREADY = 0;
        *data++ = NRF_TWI1->RXD;

        /* configure PPI to stop TWI master before we get last BB event */
        if (--data_length == 1)
        {
            if (softdevice_enabled != 0)
            {
                err_code = sd_ppi_channel_assign(s_ppi_chn, &NRF_TWI1->EVENTS_BB, &NRF_TWI1->TASKS_STOP);
                if (err_code != NRF_SUCCESS)
                {
                    return false;
                }
            }
            else
            {
                NRF_PPI->CH[s_ppi_chn].EEP = (uint32_t) &NRF_TWI1->EVENTS_BB;
                NRF_PPI->CH[s_ppi_chn].TEP = (uint32_t) &NRF_TWI1->TASKS_STOP;
            }
        }

        if (data_length == 0)
            break;
        
        nrf_delay_us(2);
        NRF_TWI1->TASKS_RESUME = 1;
    }

    /* wait until stop sequence is sent and clear the EVENTS_STOPPED */
    timeout = MAX_TIMEOUT_LOOPS;
    while(NRF_TWI1->EVENTS_STOPPED == 0 && --timeout != 0)
    {
    }
    NRF_TWI1->EVENTS_STOPPED = 0;

    if (softdevice_enabled != 0)
    {
        err_code = sd_ppi_channel_enable_clr(1 << s_ppi_chn);
        if (err_code != NRF_SUCCESS)
        {
            return err_code;
        }
    }
    else
    {
        NRF_PPI->CHENCLR = (1 << s_ppi_chn);
    }
    
    if(timeout == 0)
    {
        return false;
    }
    else
    {
        return true;
    }
}

static __inline bool attempt_recovery(void)
{
    uint32_t pselscl;
    uint32_t pselsda;
    uint32_t frequency;

    // Disable TWI
    NRF_TWI1->ENABLE = TWI_ENABLE_ENABLE_Disabled << TWI_ENABLE_ENABLE_Pos; 

    // Backup TWI settings
    pselscl = NRF_TWI1->PSELSCL;
    pselsda = NRF_TWI1->PSELSDA;
    frequency = NRF_TWI1->FREQUENCY;

    // Cycle power 
    *(uint32_t *)(NRF_TWI1_BASE + 0xFFC) = 0; 
    nrf_delay_us(5); 
    *(uint32_t *)(NRF_TWI1_BASE + 0xFFC) = 1; 
    NRF_TWI1->ENABLE = TWI_ENABLE_ENABLE_Enabled << TWI_ENABLE_ENABLE_Pos; 

    // Recover settings
    NRF_TWI1->EVENTS_RXDREADY = 0;
    NRF_TWI1->EVENTS_TXDSENT  = 0;
    NRF_TWI1->PSELSCL         = pselscl;
    NRF_TWI1->PSELSDA         = pselsda;
    NRF_TWI1->FREQUENCY       = frequency;


    return ((NRF_GPIO->IN & ((1 << s_scl_pin) | (1 << s_sda_pin))) == 
        ((1 << s_scl_pin) | (1 << s_sda_pin)));
}

/**
 * Detects stuck slaves (SDA = 0 and SCL = 1) and tries to clear the bus.
 *
 * @return
 * @retval false Bus is stuck.
 * @retval true Bus is clear.
 */
static bool hal_twi_clear_bus(void)
{
    bool bus_clear;
    
    // Disable Two-wire interface to gain GPIO access
    NRF_TWI1->ENABLE = TWI_ENABLE_ENABLE_Disabled << TWI_ENABLE_ENABLE_Pos;

    TWI_SDA_HIGH();
    TWI_SCL_HIGH();
    TWI_DELAY();

    if (TWI_SDA_READ() == 1 && TWI_SCL_READ() == 1)
    {
        bus_clear = true;
    }
    else
    {
        bus_clear = false;

        // Clock max 18 pulses worst case scenario(9 for master to send the rest of command and 9 for slave to respond) to SCL line and wait for SDA come high
        for (uint_fast8_t i=18; i--;)
        {
            TWI_SCL_LOW();
            TWI_DELAY();
            TWI_SCL_HIGH();
            TWI_DELAY();

            if (TWI_SDA_READ() == 1)
            {
                bus_clear = true;
                break;
            }
        }
    }
    
    // Re-enabling Two-wire
    NRF_TWI1->ENABLE = TWI_ENABLE_ENABLE_Enabled << TWI_ENABLE_ENABLE_Pos;;

    return bus_clear;
}

uint32_t hal_twi_init(const hal_twi_init_t* p_params)
{
    uint32_t err_code;
    uint8_t  softdevice_enabled = false;
    
    err_code = sd_softdevice_is_enabled(&softdevice_enabled);
    if (err_code != NRF_SUCCESS)
    {
        return err_code;
    }
    

    // Verifying parameters
    if ((p_params->scl_pin > 31) || 
        (p_params->sda_pin > 31) || 
        (p_params->ppi_chn > 15))
    {
        return NRF_ERROR_INVALID_PARAM;
    }
    if ((p_params->frequency != hal_twi_freq_100khz) &&
        (p_params->frequency != hal_twi_freq_250khz) &&
        (p_params->frequency != hal_twi_freq_400khz))
    {
        return NRF_ERROR_INVALID_PARAM;
    }
    
    // Checking if already initialized: If parameters this time around are the same: OK, else fail
    if ((s_scl_pin != 255) && (s_sda_pin != 255) && (s_ppi_chn != 255))
    {
        if((s_scl_pin == p_params->scl_pin) &&
           (s_sda_pin == p_params->sda_pin) &&
           (s_ppi_chn == p_params->ppi_chn) &&
           (NRF_TWI1->FREQUENCY == (p_params->frequency << TWI_FREQUENCY_FREQUENCY_Pos)))
        {
            // Already initialized, but parameters are the same
            return NRF_SUCCESS;
        }
        else
        {
            // Already initialized, but parameters are different this time
            return NRF_ERROR_BUSY;
        }
    }
    
    s_scl_pin = p_params->scl_pin;
    s_sda_pin = p_params->sda_pin;
    s_ppi_chn = p_params->ppi_chn;
        
    /* To secure correct signal levels on the pins used by the TWI
       master when the system is in OFF mode, and when the TWI master is 
       disabled, these pins must be configured in the GPIO peripheral.
    */
    NRF_GPIO->PIN_CNF[s_scl_pin] = 
        (GPIO_PIN_CNF_SENSE_Disabled << GPIO_PIN_CNF_SENSE_Pos)
      | (GPIO_PIN_CNF_DRIVE_S0D1     << GPIO_PIN_CNF_DRIVE_Pos)
      | (GPIO_PIN_CNF_PULL_Disabled  << GPIO_PIN_CNF_PULL_Pos)
      | (GPIO_PIN_CNF_INPUT_Connect  << GPIO_PIN_CNF_INPUT_Pos)
      | (GPIO_PIN_CNF_DIR_Input      << GPIO_PIN_CNF_DIR_Pos);    

    NRF_GPIO->PIN_CNF[s_sda_pin] = 
        (GPIO_PIN_CNF_SENSE_Disabled << GPIO_PIN_CNF_SENSE_Pos)
      | (GPIO_PIN_CNF_DRIVE_S0D1     << GPIO_PIN_CNF_DRIVE_Pos)
      | (GPIO_PIN_CNF_PULL_Disabled  << GPIO_PIN_CNF_PULL_Pos)
      | (GPIO_PIN_CNF_INPUT_Connect  << GPIO_PIN_CNF_INPUT_Pos)
      | (GPIO_PIN_CNF_DIR_Input      << GPIO_PIN_CNF_DIR_Pos);    

    NRF_TWI1->EVENTS_RXDREADY = 0;
    NRF_TWI1->EVENTS_TXDSENT = 0;
    NRF_TWI1->PSELSCL = s_scl_pin;
    NRF_TWI1->PSELSDA = s_sda_pin;
    NRF_TWI1->FREQUENCY = p_params->frequency << TWI_FREQUENCY_FREQUENCY_Pos;
    
    if (softdevice_enabled != 0)
    {
        err_code = sd_ppi_channel_assign(s_ppi_chn, &NRF_TWI1->EVENTS_BB, &NRF_TWI1->TASKS_SUSPEND);
        if (err_code != NRF_SUCCESS)
        {
            return err_code;
        }
        
        err_code = sd_ppi_channel_enable_clr(1 << s_ppi_chn);
        if (err_code != NRF_SUCCESS)
        {
            return err_code;
        }
    }
    else
    {
        NRF_PPI->CH[s_ppi_chn].EEP = (uint32_t) &NRF_TWI1->EVENTS_BB;
        NRF_PPI->CH[s_ppi_chn].TEP = (uint32_t) &NRF_TWI1->TASKS_SUSPEND;
        NRF_PPI->CHENCLR = (1 << s_ppi_chn);
    }
    
    NRF_TWI1->ENABLE = TWI_ENABLE_ENABLE_Enabled << TWI_ENABLE_ENABLE_Pos;

    return (hal_twi_clear_bus() == true ? NRF_SUCCESS : NRF_ERROR_INTERNAL);
}

bool hal_twi_transfer(uint8_t address, uint8_t *data, uint8_t data_length, bool issue_stop_condition)
{
    bool transfer_succeeded = true;
    
    NRF_TWI1->ENABLE = TWI_ENABLE_ENABLE_Enabled << TWI_ENABLE_ENABLE_Pos;
    
    if (data_length > 0 && hal_twi_clear_bus())
    {
        NRF_TWI1->ADDRESS = (address >> 1);

        if ((address & TWI_READ_BIT) != 0)
        {
            transfer_succeeded = hal_twi_read(data, data_length, issue_stop_condition);
        }
        else
        {
            transfer_succeeded = hal_twi_write(data, data_length, issue_stop_condition);
        }
    }
    
    NRF_TWI1->ENABLE = TWI_ENABLE_ENABLE_Disabled << TWI_ENABLE_ENABLE_Pos;
    
    if (!transfer_succeeded)
    {
        return attempt_recovery();
    }
    
    return transfer_succeeded;
}
